/* Copyright 2006 Oliver Madge
 * 
 * This file is part of RU-decoder.
 *
 * RU-decoder is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2 as published
 * by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program in the file COPYING; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301  USA
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <ctype.h>
#include <math.h>

#include <unistd.h>
#include <sys/types.h>
#include <signal.h>
#include <sys/wait.h>

#include <gtk/gtk.h>

double *probs;
int *degs;

double av, sum;

int cpid=-1;

char *k=NULL, *spares=NULL, *reps=NULL;

GtkWidget *av_label=NULL, *sum_label=NULL, *alpha_label=NULL;
GtkWidget *lambda_graph=NULL;
GtkWidget *rho_graph=NULL;
GtkWidget *text_output, *view;

GtkTextBuffer *buffer;

const int maxdegree = 8;

struct entryData {
  GtkWidget *e;
  int index;
  int array;
};

void calculate_average();
void draw_graphs();
void draw_lambda_graph();
void draw_rho_graph();

void create_top_box(GtkWidget *widget);

void calculate_alpha(GtkWidget *widget, void *);
void run_decoder(GtkWidget *widget, void *);
void kill_decoder(GtkWidget *widget, void *);
void update_text(GtkWidget *widget, char*);

void calculate_alpha(GtkWidget *widget, void *v) {
  printf("Not implemented yet\n");
}

//void run_decoder(GtkWidget *widget, GtkWidget *kill) {
void run_decoder(GtkWidget *widget, void *kill) {
  //char *buf;
  char buf;
  char *degS=(char *)malloc(sizeof(char)*(20*maxdegree+1));
  degS[0]='\0';
  timeval tv;
  fd_set rfds;
  GtkTextIter iter;
  int pfd[2];
  if(cpid!=-1) {
    printf("Decoder is already running\n");
    return;
  }

  //grey out "Run decoder" button
  gtk_widget_set_sensitive(widget, FALSE);
  gtk_widget_set_sensitive(GTK_WIDGET(kill), TRUE);

  gtk_text_buffer_get_end_iter(buffer, &iter);
  GtkTextMark* end_mark = gtk_text_buffer_create_mark(buffer, NULL, &iter, FALSE);

  //build degree distribution into degS
  char *tmpptr = degS;
  for(int i=0; i<maxdegree; i++) {
    if(degs[i]!=0) {
      int ret = snprintf(tmpptr, 20, "(%i,%f);", degs[i], probs[i]);
      tmpptr+=ret;
      //printf("%s\n", degS);
    }
  }

  printf("%s\n", degS);
  
    
  //buf = (char *)malloc(sizeof(char)*100);
  printf("Running decoder\n");
    
  if (pipe(pfd) == -1) { perror("pipe"); return; }
  cpid = fork();
  if(cpid==0) {
    close(pfd[0]);
    dup2(pfd[1], fileno(stdout));
    //execl("../sparse/decoder", "%s %s %s %s", k, "1", spares, "(2,0.4252);(3,0.3552);(6,0.0600);(15,0.1400);(300,0.01959);", (char *) NULL);
    execl("../sparse/decoder", "%s %s %s %s %s", k, "1", spares, degS, reps, (char *) NULL);
    cpid=-1;
  }
  else {
    close(pfd[1]);

    int retval;
    buf = 'a';

    GtkTextIter start, end;
    gtk_text_buffer_get_start_iter(buffer, &start);
    gtk_text_buffer_get_end_iter(buffer, &end);
    gtk_text_buffer_delete(GTK_TEXT_BUFFER(buffer), &start, &end);
    
    //TODO: append degree distribution and decoder output to logfile
    
    while(buf!=EOF) {
      tv.tv_sec=0;
      FD_ZERO(&rfds);
      FD_SET(pfd[0], &rfds);     tv.tv_usec=20;
      retval = select(pfd[0]+1, &rfds, NULL, NULL, &tv);
      if(retval == -1) { printf("Error\n"); }
      else if(retval) {
        //printf("Data\n");
        if(read(pfd[0], &buf, 1)<=0) break;
        
        gtk_text_buffer_get_end_iter(buffer, &iter);
        gtk_text_buffer_insert(GTK_TEXT_BUFFER(buffer), &iter, &buf, 1);
        gtk_text_view_scroll_to_mark(GTK_TEXT_VIEW(view), end_mark, 0.0, FALSE, 0.0, 0.0);
      }
      else {
        //printf("No data\n");
        //do a processing loop so the GUI stays responsive
        gtk_main_iteration_do(FALSE);
      }
    
    }

    wait(NULL);
    cpid=-1;

  }
  gtk_widget_set_sensitive(GTK_WIDGET(widget), TRUE);
  gtk_widget_set_sensitive(GTK_WIDGET(kill), FALSE);
  free(degS);
  //system(buf);
}

void kill_decoder(GtkWidget *widget, void *v) {
  if(cpid!=-1) {
    kill(cpid, SIGTERM);
    cpid=-1;
  }
  else printf("No decoder process running\n");
}

void update_text(GtkWidget *widget, char *buf) {
  const gchar *entry_text;
  entry_text = gtk_entry_get_text(GTK_ENTRY(widget));
  strncpy(buf, entry_text, 100);
  buf[99]=0;
}

static GdkPixmap *pixmap = NULL, *pixmap2 = NULL;
/* Create a new backing pixmap of the appropriate size */
static gboolean configure_event_lambda(GtkWidget *widget, GdkEventConfigure *event) {
  if(pixmap) g_object_unref(pixmap);
  pixmap = gdk_pixmap_new(widget->window, widget->allocation.width, widget->allocation.height, -1);
  gdk_draw_rectangle (pixmap, widget->style->white_gc, TRUE, 0, 0, widget->allocation.width, widget->allocation.height);
  draw_lambda_graph();
  return TRUE;
}

static gboolean configure_event_rho(GtkWidget *widget, GdkEventConfigure *event) {
  if(pixmap2) g_object_unref(pixmap2);
  pixmap2 = gdk_pixmap_new(widget->window, widget->allocation.width, widget->allocation.height, -1);
  gdk_draw_rectangle (pixmap2, widget->style->light_gc[1], TRUE, 0, 0, widget->allocation.width, widget->allocation.height);
  draw_rho_graph();
  return TRUE;
}

static gboolean expose_event_lambda(GtkWidget *widget, GdkEventExpose *event) {
  gdk_draw_drawable(widget->window, widget->style->fg_gc[GTK_WIDGET_STATE (widget)], pixmap, event->area.x, event->area.y, event->area.x, event->area.y, event->area.width, event->area.height);
  return FALSE;
}

static gboolean expose_event_rho(GtkWidget *widget, GdkEventExpose *event) {
  gdk_draw_drawable(widget->window, widget->style->fg_gc[GTK_WIDGET_STATE (widget)], pixmap2, event->area.x, event->area.y, event->area.x, event->area.y, event->area.width, event->area.height);
  return FALSE;
}


static void draw_brush (GtkWidget *widget, gint x, gint y, gint w, gint h) {
  GdkRectangle update_rect;
  update_rect.x = x;
  update_rect.y = y;
  update_rect.width = w;
  update_rect.height = h;
  gdk_draw_rectangle (pixmap, widget->style->black_gc, TRUE, update_rect.x, update_rect.y, update_rect.width, update_rect.height);
  gtk_widget_queue_draw_area (widget, update_rect.x, update_rect.y, update_rect.width, update_rect.height);
}

static void draw_brush_rho(GtkWidget *widget, gint x, gint y, gint w, gint h) {
  GdkRectangle update_rect;
  update_rect.x = x;
  update_rect.y = y;
  update_rect.width = w;
  update_rect.height = h;
  gdk_draw_rectangle (pixmap2, widget->style->black_gc, TRUE, update_rect.x, update_rect.y, update_rect.width, update_rect.height);
  gtk_widget_queue_draw_area (widget, update_rect.x, update_rect.y, update_rect.width, update_rect.height);
}

static void clear_screen(GtkWidget *widget) {
  gdk_draw_rectangle(pixmap, widget->style->bg_gc[0], TRUE, 0, 0, widget->allocation.width, widget->allocation.height);
  gtk_widget_queue_draw_area(widget, 0, 0, widget->allocation.width, widget->allocation.height);
}

static void clear_screen_rho(GtkWidget *widget) {
  gdk_draw_rectangle(pixmap2, widget->style->bg_gc[0], TRUE, 0, 0, widget->allocation.width, widget->allocation.height);
  gtk_widget_queue_draw_area(widget, 0, 0, widget->allocation.width, widget->allocation.height);
}

static gboolean delete_event( GtkWidget *widget, GdkEvent  *event, gpointer   data) {
  gtk_main_quit();
  return FALSE;
}

static void text_validate_numerical(GtkWidget *widget, gchar *new_text, gint new_text_length, gint *position, entryData *entry) {
  //TODO: needs some fixing
  const gchar *entry_text;
  entry_text = gtk_entry_get_text(GTK_ENTRY(entry->e));
  //printf("%i %s %s\n", *position, new_text, entry_text);
  int start=0;
  switch(entry->array) {
    case 1:
      //allow .
      if(*position==0 && new_text[0]=='.' && entry_text[0]!='.' && entry_text[1]!='.') return;
      if(*position==1 && new_text[0]=='.' && entry_text[0]=='0' && entry_text[1]!='.') return;
      if(*position==0  && (new_text[0]!='.' && new_text[0]!='0')) {
        g_signal_stop_emission_by_name(widget, "insert-text");
        return;
      }
      if(*position==1  && (new_text[0]!='.' || new_text[0]!='0')) {
        return;
      }
      start=2;
    case 0:
      //if(*new_text=='0' || *new_text=='1' || *new_text=='2' ||*new_text=='3' ||*new_text=='4') return;
      //if(*new_text=='5' || *new_text=='6' ||*new_text=='7' || *new_text=='8' || *new_text=='9') return;
      
      for(int i=start; i<new_text_length; i++) {
        if(!isdigit(new_text[i])) {
          g_signal_stop_emission_by_name(widget, "insert-text");
          return;
        }
      }
  }
}

static void text_changed_callback(GtkWidget *widget, entryData *entry) {
  const gchar *entry_text;
  entry_text = gtk_entry_get_text(GTK_ENTRY(entry->e));
  
  switch(entry->array) {
    case 0:
      //TODO: check entry_text is an integer
      degs[entry->index] = atoi(entry_text);
      //printf ("Entry contents %i: %s\n",entry->index, entry_text);
      break;
    case 1:
      probs[entry->index] = atof(entry_text);
      //printf ("Entry contents %i: %s\n",entry->index, entry_text);
  }
  calculate_average();
  draw_graphs();
}

void calculate_average() {
  av=0.0, sum=0.0;
  gchar *av_text = (char*)malloc(sizeof(char)*100);
  gchar *sum_text = (char*)malloc(sizeof(char)*100);
  for(int i=0; i<maxdegree; i++) {
    av+=degs[i]*probs[i];
    sum+=probs[i];
  }
  sprintf(av_text, "%e", av);
  sprintf(sum_text, "%e", sum);
  if(av_label!=NULL) gtk_label_set_text(GTK_LABEL(av_label), av_text);
  if(sum_label!=NULL) gtk_label_set_text(GTK_LABEL(sum_label), sum_text);
  free(av_text); free(sum_text);
}

void draw_graphs() {
  if(lambda_graph!=NULL) draw_lambda_graph();
  if(rho_graph!=NULL) draw_rho_graph();
}

void draw_lambda_graph() {
  clear_screen(lambda_graph);
  gint w = lambda_graph->allocation.width;
  gint h = lambda_graph->allocation.height;
  
  //draw axes
  //x axis
  draw_brush(lambda_graph, 2, 12 , 2, h-15 );
  //y axis
  draw_brush(lambda_graph, 2, h-5 , w-15, 2 );
  gint scale = 2;
  double curmax=0.0;
  for(int i=0; i<maxdegree; i++) if(degs[i]!=0 && probs[i]>curmax) curmax = probs[i];
  
  scale = (gint)floor(1/curmax);
  for(int i=0; i<maxdegree; i++) {
    if(probs[i]!=0.0 && degs[i]!=0) draw_brush(lambda_graph, degs[i]*3, (gint) (12 + (h-15) * (1 - scale*probs[i])), 2, (gint)((h-15)*probs[i]*scale ));
  }
}

void draw_rho_graph() {
  clear_screen_rho(rho_graph);
  gint w = rho_graph->allocation.width;
  gint h = rho_graph->allocation.height;
  gint scale=2;
  //draw axes
  //x axis
  draw_brush_rho(rho_graph, 2, 12 , 2, h-15 );
  //y axis
  draw_brush_rho(rho_graph, 2, h-5 , w-15, 2 );
  if(av!=0.0) {
    double cur = exp(-av), curmax=0.0;
    for(int i=1; i<150; i++) {
      cur=cur * av / i;
      if(cur>curmax) curmax = cur;
    }
    cur = exp(-av);
    scale = (gint)floor(1/curmax);
    for(int i=1; i<150; i++) {
      cur=cur * av / i;
      draw_brush_rho(rho_graph, i*3, (gint)(12 + (h-15) * (1 - scale*cur)), 2, (gint)((h-15)*cur*scale) );
    }
  }
}

int main(int argc, char **argv) {

  GtkWidget *window;
  GtkWidget *vbox;
  GtkWidget *graph_hbox;
  GtkWidget *button;


  GtkWidget *hsep;
  GtkWidget *frame;
  

  gtk_init(&argc, &argv);

  //allocate memory for probability and degrees
  
  probs = (double*)malloc(sizeof(double)*maxdegree);
  degs = (int*)malloc(sizeof(int)*maxdegree);
  for(int i=0; i<maxdegree; i++) {
    probs[i]=0.0; degs[i]=0;
  }

  degs[0]=2;
  probs[0]=0.4252;
  degs[1]=3;
  probs[1]=0.3552;
  degs[2]=7;
  probs[2]=0.0600;
  degs[3]=14;
  probs[3]=0.0700;
  degs[4]=15;
  probs[4]=0.0700;
  degs[5]=350;
  probs[5]=0.01960;
  
  k = (char*)malloc(sizeof(char)*100);
  spares = (char*)malloc(sizeof(char)*100);
  reps = (char*)malloc(sizeof(char)*100);
  k[0]='\0'; spares[0]='\0'; reps[0]='\0';
    
  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_container_set_border_width(GTK_CONTAINER (window), 10);
  gtk_window_set_title(GTK_WINDOW(window), "Degree distribution analyser");
  g_signal_connect(G_OBJECT(window), "delete_event", G_CALLBACK (delete_event), NULL);

  vbox = gtk_vbox_new(FALSE, 10);
  //gtk_container_set_border_width(GTK_CONTAINER(vbox), 10);

  gtk_container_add(GTK_CONTAINER(window), vbox);
 
  create_top_box(vbox);

  hsep = gtk_hseparator_new();
  gtk_box_pack_start(GTK_BOX(vbox), hsep, TRUE, TRUE, 0);
  gtk_widget_show(hsep);
 
  graph_hbox = gtk_hbox_new(FALSE, 40);
  gtk_box_pack_start(GTK_BOX(vbox), graph_hbox, TRUE, TRUE, 0);
  
  frame = gtk_frame_new("Lambda");
  gtk_box_pack_start(GTK_BOX(graph_hbox), frame, TRUE, TRUE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (frame), 10);

  lambda_graph = gtk_drawing_area_new();
  gtk_signal_connect (GTK_OBJECT (lambda_graph), "expose_event",
          (GtkSignalFunc) expose_event_lambda, NULL);
  gtk_signal_connect (GTK_OBJECT(lambda_graph),"configure_event",
          (GtkSignalFunc) configure_event_lambda, NULL);
  gtk_widget_set_size_request(lambda_graph, 315, 315);

  gtk_widget_set_style(lambda_graph, gtk_widget_get_style(frame));
  
  //gtk_box_pack_start(GTK_FRAMEgtk_container_add (GTK_CONTAINER (window), frame);
  //    (frame), lambda_graph, TRUE, TRUE, 0);
  gtk_container_add (GTK_CONTAINER (frame), lambda_graph);
  gtk_widget_show(lambda_graph);

  gtk_widget_show(frame);

  frame = gtk_frame_new("Rho");
  gtk_box_pack_start(GTK_BOX(graph_hbox), frame, TRUE, TRUE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (frame), 10);

  rho_graph = gtk_drawing_area_new();
  gtk_signal_connect (GTK_OBJECT (rho_graph), "expose_event",
          (GtkSignalFunc) expose_event_rho, NULL);
  gtk_signal_connect (GTK_OBJECT(rho_graph),"configure_event",
          (GtkSignalFunc) configure_event_rho, NULL);
  gtk_widget_set_size_request(rho_graph, 315, 315);
  
  gtk_container_add (GTK_CONTAINER (frame), rho_graph);
  
  gtk_widget_show(rho_graph);

  gtk_widget_show(frame);

  gtk_widget_show(graph_hbox);

  gtk_widget_show(vbox);
  printf("%i, %i\n", gdk_screen_width(), gdk_screen_height());
  gtk_window_move(GTK_WINDOW(window), gdk_screen_width()/3,gdk_screen_height()/3);
  gtk_widget_show(window);
    
  gtk_main (); 

  free(probs); free(degs);
  free(k); free(spares);
  free(reps);
  return 0;
}

void create_top_box(GtkWidget *vbox) {
  GtkWidget *table;
  GtkWidget *top_hbox, *right_vbox;

  GtkWidget *label;

  GtkWidget *entry;
  GtkWidget *vsep;
  GtkWidget *button,*button2;

  //GtkWidget *view;
  GtkWidget *scroll;


  top_hbox = gtk_hbox_new(FALSE, 40);
  gtk_box_pack_start(GTK_BOX(vbox), top_hbox, TRUE, TRUE, 0);

  /* Top left hand table for lambda input */
  
  table = gtk_table_new(maxdegree+1, 2, TRUE);
  gtk_table_set_row_spacings(GTK_TABLE(table), 5);
  gtk_table_set_col_spacings(GTK_TABLE(table), 45);
  gtk_box_pack_start(GTK_BOX(top_hbox), table, FALSE, TRUE, 0);

  label = gtk_label_new ("Degree");
  gtk_table_attach_defaults(GTK_TABLE(table), label, 0, 1, 0, 1);
  gtk_widget_show(label);
 
  label = gtk_label_new ("Probability");
  gtk_table_attach_defaults (GTK_TABLE(table), label, 1, 2, 0, 1);
  gtk_widget_show(label);

  char *tmpBuff = (char *)malloc(sizeof(char)*(20+1));
  for(int j=0; j<2; j++) {
    for(int i=0; i<maxdegree; i++) {
      entry = gtk_entry_new();
      entryData *data = (entryData*)g_malloc(sizeof(entryData));
      data->index=i; data->e=entry; data->array=j;
      
      g_signal_connect(G_OBJECT(entry), "changed", G_CALLBACK(text_changed_callback), data);
      g_signal_connect(G_OBJECT(entry), "insert-text", G_CALLBACK(text_validate_numerical), data);
      g_signal_connect_swapped(G_OBJECT(entry), "destroy", G_CALLBACK(g_free), data);
      gtk_entry_set_max_length(GTK_ENTRY(entry), 10);
      gtk_table_attach_defaults(GTK_TABLE(table), entry, j, j+1, i+1, i+2);
      switch(j) {
        case 0:
          if(degs[i]!=0) {
            snprintf(tmpBuff, 20, "%i", degs[i]);
            tmpBuff[20]='\0';
            gtk_entry_set_text(GTK_ENTRY(entry), tmpBuff);
          }
          break;
        case 1:
          if(probs[i]!=0.0) {
            snprintf(tmpBuff, 20, "%f", probs[i]);
            tmpBuff[20]='\0';
            gtk_entry_set_text(GTK_ENTRY(entry), tmpBuff);
          }
      }
      gtk_widget_show(entry);
    }
  }
  free(tmpBuff);

  gtk_widget_show(table);

  vsep = gtk_vseparator_new();
  gtk_box_pack_start(GTK_BOX(top_hbox), vsep, FALSE, TRUE, 0);
  gtk_widget_show(vsep);

  right_vbox = gtk_vbox_new(FALSE, 10);
  gtk_box_pack_start(GTK_BOX(top_hbox), right_vbox, FALSE, FALSE, 10);
  /* Top right hand table for information display */
  
  table = gtk_table_new(3, 2, TRUE);
  gtk_table_set_row_spacings(GTK_TABLE(table), 25);
  gtk_table_set_col_spacings(GTK_TABLE(table), 15);
  gtk_box_pack_start(GTK_BOX(right_vbox), table, FALSE, FALSE, 0);


  label = gtk_label_new("Average:");
  gtk_table_attach(GTK_TABLE(table), label, 0, 1, 0, 1, (GtkAttachOptions)0, (GtkAttachOptions)0, 0, 0);
  //gtk_table_attach_defaults(GTK_TABLE(table), label, 0, 1, 0, 1);
  gtk_widget_show(label);
  
  label = gtk_label_new("ln(K):");
  gtk_table_attach(GTK_TABLE(table), label, 0, 1, 1, 2, (GtkAttachOptions)0, (GtkAttachOptions)0, 0, 0);
  //gtk_table_attach_defaults(GTK_TABLE(table), label, 0, 1, 0, 1);
  gtk_widget_show(label);

  label = gtk_label_new("Sum:");
  gtk_table_attach(GTK_TABLE(table), label, 0, 1, 2, 3, (GtkAttachOptions)0, (GtkAttachOptions)0, 0, 0);
  //gtk_table_attach_defaults(GTK_TABLE(table), label, 0, 1, 1, 2);
  gtk_widget_show(label);
  
  label = gtk_label_new("\\alpha*:");
  gtk_table_attach(GTK_TABLE(table), label, 0, 1, 3, 4, (GtkAttachOptions)0, (GtkAttachOptions)0, 0, 0);
  //gtk_table_attach_defaults(GTK_TABLE(table), label, 0, 1, 1, 2);
  gtk_widget_show(label);

  av_label = gtk_label_new("Average");
  gtk_table_attach(GTK_TABLE(table), av_label, 1, 2, 0, 1, (GtkAttachOptions)0, (GtkAttachOptions)0, 0, 0);
  //gtk_table_attach_defaults(GTK_TABLE(table), av_label, 1, 2, 0, 1);
  gtk_widget_show(av_label);
  
  label = gtk_label_new("9.21034");
  gtk_table_attach(GTK_TABLE(table), label, 1, 2, 1, 2, (GtkAttachOptions)0, (GtkAttachOptions)0, 0, 0);
  //gtk_table_attach_defaults(GTK_TABLE(table), sum_label, 1, 2, 1, 2);
  gtk_widget_show(label);
  
  sum_label = gtk_label_new("Sum");
  gtk_table_attach(GTK_TABLE(table), sum_label, 1, 2, 2, 3, (GtkAttachOptions)0, (GtkAttachOptions)0, 0, 0);
  //gtk_table_attach_defaults(GTK_TABLE(table), sum_label, 1, 2, 1, 2);
  gtk_widget_show(sum_label);

  alpha_label = gtk_label_new("Alpha");
  gtk_table_attach(GTK_TABLE(table), alpha_label, 1, 2, 3, 4, (GtkAttachOptions)0, (GtkAttachOptions)0, 0, 0);
  //gtk_table_attach_defaults(GTK_TABLE(table), sum_label, 1, 2, 1, 2);
  gtk_widget_show(alpha_label);
  
  calculate_average();

  gtk_widget_show(table);

  button = gtk_button_new_with_label("Calculate Alpha");
  g_signal_connect(G_OBJECT(button), "clicked", G_CALLBACK(calculate_alpha), NULL);
  gtk_box_pack_start(GTK_BOX(right_vbox), button, FALSE, FALSE, 10);
  gtk_widget_show(button);

  table = gtk_table_new(3, 2, FALSE);
  gtk_table_set_row_spacings(GTK_TABLE(table), 5);
  gtk_table_set_col_spacings(GTK_TABLE(table), 45);
  gtk_box_pack_start(GTK_BOX(right_vbox), table, FALSE, TRUE, 0);

  label = gtk_label_new("K:");
  gtk_table_attach(GTK_TABLE(table), label, 0, 1, 0, 1, (GtkAttachOptions)0, (GtkAttachOptions)0, 0, 0);
  gtk_widget_show(label);

  label = gtk_label_new("Spares:");
  gtk_table_attach(GTK_TABLE(table), label, 0, 1, 1, 2, (GtkAttachOptions)0, (GtkAttachOptions)0, 0, 0);
  gtk_widget_show(label);

  label = gtk_label_new("Reps:");
  gtk_table_attach(GTK_TABLE(table), label, 0, 1, 2, 3, (GtkAttachOptions)0, (GtkAttachOptions)0, 0, 0);
  gtk_widget_show(label);

  entry = gtk_entry_new();
  gtk_table_attach(GTK_TABLE(table), entry, 1, 2, 0, 1, (GtkAttachOptions)0, (GtkAttachOptions)0, 0, 0);
  g_signal_connect(G_OBJECT(entry), "changed", G_CALLBACK(update_text), k);
  gtk_entry_set_text(GTK_ENTRY(entry), "10000");
  gtk_widget_show(entry);
  
  entry = gtk_entry_new();
  gtk_table_attach(GTK_TABLE(table), entry, 1, 2, 1, 2, (GtkAttachOptions)0, (GtkAttachOptions)0, 0, 0);
  g_signal_connect(G_OBJECT(entry), "changed", G_CALLBACK(update_text), spares);
  gtk_entry_set_text(GTK_ENTRY(entry), "200");
  gtk_widget_show(entry);

  entry = gtk_entry_new();
  gtk_table_attach(GTK_TABLE(table), entry, 1, 2, 2, 3, (GtkAttachOptions)0, (GtkAttachOptions)0, 0, 0);
  g_signal_connect(G_OBJECT(entry), "changed", G_CALLBACK(update_text), reps);
  gtk_entry_set_text(GTK_ENTRY(entry), "50");
  gtk_widget_show(entry);

  gtk_widget_show(table);

  button2 = gtk_button_new_with_label("Kill Decoder");
  button = gtk_button_new_with_label("Run Decoder");
  g_signal_connect(G_OBJECT(button), "clicked", G_CALLBACK(run_decoder), button2);
  gtk_box_pack_start(GTK_BOX(right_vbox), button, FALSE, FALSE, 0);
  gtk_widget_show(button);

  g_signal_connect(G_OBJECT(button2), "clicked", G_CALLBACK(kill_decoder), NULL);
  gtk_box_pack_start(GTK_BOX(right_vbox), button2, FALSE, FALSE, 0);
  gtk_widget_set_sensitive(button2, FALSE);
  gtk_widget_show(button2);

  gtk_widget_show(right_vbox);

  
  scroll = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scroll), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

  view = gtk_text_view_new();
  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(view));
  gtk_text_view_set_editable(GTK_TEXT_VIEW(view), FALSE);
  gtk_text_buffer_set_text(buffer, "", -1); 
  gtk_container_add(GTK_CONTAINER(scroll), view);
  gtk_widget_set_size_request(view, 200, 315);
  gtk_widget_show(view);
  
  gtk_box_pack_start(GTK_BOX(top_hbox), scroll, FALSE, FALSE, 0);
  gtk_widget_show(scroll);
  
  gtk_widget_show(top_hbox);



}

