/* Copyright 2006 Oliver Madge
 * 
 * This file is part of RU-decoder.
 *
 * RU-decoder is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2 as published
 * by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program in the file COPYING; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301  USA
 *
 */

#include "columnvector.h"

columnVector::columnVector(int h){
  data = new int[h];
  this->height = h;
  this->clear();
}

columnVector::~columnVector(){
  delete [] data;
}

void columnVector::clear() {
  for(int i=0; i<this->height; i++) {
    this->data[i]=0;
  }
}

void columnVector::add(const columnVector *input){
  for(int i=0; i<this->height; i++) {
    this->data[i] += input->data[i];
  }
}

void columnVector::bitwiseAdd(const columnVector *input){
  for(int i=0; i<this->height; i++) {
    this->data[i] ^= input->data[i];
  }
}

void columnVector::bitwiseAdd(const columnVector *input, int offset, int length){
  for(int i=offset; i<offset+length; i++) {
    this->data[i] ^= input->data[i-offset];
  }
}

void columnVector::subtract(const columnVector *input){
  for(int i=0; i<this->height; i++) {
    this->data[i] -= input->data[i];
  }
}

void columnVector::copy(const columnVector *input){
  //TODO: check we are the same size
  for(int i=0; i<this->height; i++) {
    this->data[i] = input->data[i];
  }
}

int columnVector::get(int index){
  if(!inRange(index)) {
    printf("Requested subscript %i out of range\n", index);
    return -1;
  }
  return data[index];
}

void columnVector::push(int index, int value){
  if(!inRange(index)) {
    //printf("Out of range\n");
    return; 
  }
  data[index] = value;
}

void columnVector::swapRows(int i, int j){
  if(!inRange(i)) { printf("i out of range, (%i, %i)\n", i,j); return; }
  if(!inRange(j)) { printf("j out of range\n"); return; }
  int temp = data[i];
  data[i] = data[j];
  data[j] = temp;
}


void columnVector::prettyPrint(){
  for(int i=0; i<this->height; i++) {
    printf("|%i|\n", data[i]);
  }
}

void columnVector::prettyPrintTranspose(){
  printf("| ");
  for(int i=0; i<this->height; i++) {
    printf("%i ", data[i]);
  }
  printf("|\n");
}

//ugly and slow
void columnVector::prettyPrintTransposeSort(){
  printf("| ");
  int min, max=0;
  for(int j=0; j<this->height; j++) {
    min=999999;
    for(int i=0; i<this->height; i++) {
      if(data[i]<min && data[i]>=max) min=data[i];
    }
    printf("%i ", min);
    max=min+1;
  }
  printf("|\n");
}

bool columnVector::inRange(int index) {
  if(index<0 || index>=this->height) return false;
  return true;
}
