/* Copyright 2006 Oliver Madge
 * 
 * This file is part of RU-decoder.
 *
 * RU-decoder is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2 as published
 * by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program in the file COPYING; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301  USA
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "sparse.h"
#include "dense.h"
#include "columnvector.h"
#include "degree.h"

int size;
int spareSize;

void printCounts(sparseMatrix *H) {
  debugPrint("-- %i ",H->getBinaryOpCount());
  debugPrint("%i ",H->getSymbolOpCount());
  debugPrint("%i\n",H->getSwapOpCount());
}

void receiveMessage(degree *dDist, sparseMatrix *H, columnVector *r) {
  int tempI, binaryOps=0, symbolOps=0, swapOps=0;
  list<int> *plist = NULL;
  list<int>::iterator iter;

  int *rWeights = (int *)malloc(sizeof(int)*size);

  for(int i=0; i<size; i++) rWeights[i]=0;


  debugPrint("Receiving message\n");
  //"receive" message
  for(int i=0; i<size+spareSize; i++) {
    uint16_t deg = dDist->getDegree();
    rWeights[deg]++;
    if(deg==0) printf("Error, degree zero packet\n");
    plist = dDist->getPacketList(deg, size);
    //printf("%i, %i\n", deg, plist->size());

    H->insertRow(i, plist);
    

    tempI=0;
    iter = plist->begin();
    while(iter!=plist->end()){
      tempI^=((*iter)+'a');
      iter++;
    }
    
    r->push(i, tempI);
    symbolOps++;
  }


  //FIXME: hack. Sparse matrix should keep track itself
  H->copyColumnsFromRows();

  debugPrint("-- %i ", H->getBinaryOpCount());
  debugPrint("%i ", symbolOps);
  debugPrint("%i\n", swapOps);

  //for(int i=0; i<size; i++) if(rWeights[i]!=0) printf("-- %i %f\n", i, (double)((double)rWeights[i]/(double)(size+spareSize)));
  free(rWeights);
}

//(2,0.4252);(3,0.3552);(6,0.0600);(15,0.1400);(500,0.01959)

//FIXME: there's a bug in here somewhere -- happy hunting!
void parseDegreeDist(const char *in, int **degi, double **degd){
  char *w = (char *)malloc(sizeof(char)*(strlen(in)+1));
  char *buff = (char *)malloc(sizeof(char)*(strlen(in)+1));
  strcpy(w, in);

  int l = 1;

  *degi = (int *)malloc(sizeof(int)*l);
  *degd = (double *)malloc(sizeof(double)*l);

  //don't find the number of ;, resize arrays as we go
  
  char *rw=w+1; //skip trailing (
  int offset;

  while(*rw!='\0') {
    offset=strcspn(rw, ";");
    if(offset<=1) break;
    strncpy(buff, rw, offset);
    buff[offset]='\0';

    l++;
    *degi = (int *)realloc(*degi, sizeof(int)*l);
    *degd = (double *)realloc(*degd, sizeof(double)*l);

    char *offset2 = index(buff, ',');
    *offset2 = '\0';
    (*degi)[l-2] = atoi(buff);   //first bit of buff
    
    offset2++;
    char *offset3 = index(offset2, ')');
    *offset3 = '\0';

    (*degd)[l-2] = atof(offset2); //second bit of buff

    rw = rw + offset+1+1; //skip ( too
  }

  (*degi)[l-1]=0;

  free(buff); free(w);

}


int invert(sparseMatrix *H, columnVector *p, columnVector *r, columnVector *ans) {

  //greedy search of largest t
  //start with first row, look at 2nd element, record column position

  //TODO: examine the usefulness of this function
  //      does it ever differ from the sizeOfT returned by triangulate?
  //      if not we can remove it
  int i=1, maximum=H->getWidth(), ret=0;
  while(i<=maximum) { //while we're not in the "maximum" row, t must be square
    //look at row, check for diagonal and next element, if less than previous
    //update previous
    if(H->getFirstRowValue(i)!=i) break;
    i++;
  }

  //printf("size of triangle: %i\n", i-1);

  int heightOfT = i-1;

  int g = H->getWidth()-heightOfT;

  columnVector *p2 = new columnVector(heightOfT);
  columnVector *p1 = new columnVector(g);

  if(g==0) {  //special case where we have a perfect triangle
    H->pivotEliminate(r, p);
  }
  else
  {
    //2. separate h into t, e, b and d
    sparseMatrix *t=new sparseMatrix(heightOfT,heightOfT);
    sparseMatrix *b=new sparseMatrix(g, heightOfT);
    sparseMatrix *e=new sparseMatrix(heightOfT, g + spareSize);
    sparseMatrix *d=new sparseMatrix(g, g + spareSize);

    debugPrint("Splitting matrices\n");

    H->clearOpCounts();

    H->split(1,1,heightOfT,heightOfT,t);
    H->split(heightOfT+1,1, g,heightOfT,b);
    H->split(1,heightOfT+1, heightOfT, g+spareSize, e);
    H->split(heightOfT+1,heightOfT+1, g, g+spareSize, d);

    printCounts(H);
    printCounts(t);
    printCounts(b);
    printCounts(e);
    printCounts(d);

    //separate r into r_a (of height t) and r_b (the rest of r (height g+spareSize))
    debugPrint("Splitting vectors\n");
    
    columnVector *r_a = new columnVector(heightOfT);
    columnVector *r_b = new columnVector(g+spareSize);

    for(int i=0; i<heightOfT; i++) r_a->push(i,r->get(i));
    for(int i=0; i<g+spareSize; i++) r_b->push(i,r->get(heightOfT+i));

    debugPrint("-- 0 %i 0\n", size+spareSize);

    //3. calculate p2_a = t{-1} * r_a (pivot elimination)

    t->clearOpCounts();
    debugPrint("Calculating p2_a\n");

    columnVector *p2_a = new columnVector(heightOfT);
    t->pivotEliminate2(r_a, p2_a);

    printCounts(t);

    //printf("p2_a=\n");
    //p2_a->prettyPrintTranspose();

    //4. calculate z_b = r_b - e * p2_a

    e->clearOpCounts();
    debugPrint("Calculating z_b\n");  

    columnVector *z_b = new columnVector(g+spareSize);

    e->preMultiply(p2_a, z_b);

    z_b->bitwiseAdd(r_b);

    delete p2_a;
    
    printCounts(e);
    debugPrint("-- 0 %i 0\n", r_b->getHeight());

    //printf("r_b=\n");
    //r_b->prettyPrintTranspose();

    //5. calculate f = - e * t{-1} * b + d (pivot elimination again on each column of b)

    sparseMatrix *f = new sparseMatrix(g, g+spareSize);

    t->clearOpCounts();
    debugPrint("Calculating f"); fflush(stdout);

    sparseMatrix *temp2 = new sparseMatrix(g, heightOfT);

    //printf("b=\n");
    //b->prettyPrint();

    debugPrint("."); fflush(stdout);
    t->pivotEliminate2(b, temp2);

    printCounts(t);
    printCounts(temp2);

    //printf("temp2=\n");
    //temp2->prettyPrint();

    debugPrint("."); fflush(stdout);

    e->clearOpCounts();
    f->clearOpCounts();

    e->multiply(temp2, f);
    
    printCounts(e);
    printCounts(f);

    delete temp2;

    f->clearOpCounts();

    debugPrint(".\n");
    f->add(d);

    printCounts(f);

    //6. p1 = f{-1} * z_b

    f->clearOpCounts();
    debugPrint("Inverting f\n");
    ret = f->slackInvert(z_b, p1); //p1 returned in correct order
    //if(ret==-1) exit(-1);
    printCounts(f);

    delete z_b;

    //7. calculate z_c = r_a + b * p1

    b->clearOpCounts();
    debugPrint("Calculating z_c\n");
    columnVector *z_c = new columnVector(heightOfT);

    b->preMultiply(p1, z_c);

    z_c->bitwiseAdd(r_a);

    printCounts(b);
    debugPrint("-- 0 %i 0\n", z_c->getHeight());


    //8. p2 = t{-1} * z_c (pivot elimination again)
    
    t->clearOpCounts();
    debugPrint("%i spares used\nCalculating p2\n", ret);
    
    t->pivotEliminate2(z_c, p2);

    printCounts(t);

    delete z_c;
     delete t;
    delete e;  delete b;  delete d;
    delete r_a; delete r_b;

    delete f;
    
    //copy p2 to top of p
    p->bitwiseAdd(p2, 0, heightOfT);
    //copy p1 to bottow of p
    p->bitwiseAdd(p1, heightOfT, g);
  }
  delete p2;  delete p1;

  return ret;
}


int main(int argc, char **argv){

  if(argc!=6){
    printf("decoder size seed spares degdist reps\n");
    return -1;
  }

  /*int c;
  while((c=getopt(argc, argv, "v"))!=-1) {
    printf("Option %c\n", c);
    switch(c) {
      case 'v':
        printf("Verbose mode\n");
      default:
        ;
    }
  }*/

  
  //TODO: some error handling here (OK; a lot)
  size=atoi(argv[1]);
  spareSize=atoi(argv[3]);

  int reps = atoi(argv[5]);

  int seed = atoi(argv[2]);

  unsetDebugMode();

  int avTri=0, reps2 = reps, failed=0, fail2=0, blankCols=0;

  int minTri = size, maxTri=0;

  int *degi;
  double *degd;
  parseDegreeDist(argv[4], &degi, &degd);

  while(reps>0) {
    reps--;

  Random *rDist = new Random();
  rDist->RandomInit(seed);


  degree *dDist = new degree(degi, degd, size, rDist);


  //double alpha = dDist->alpha_star();

  sparseMatrix *H = new sparseMatrix(size, 0);
  columnVector *r = new columnVector(size+spareSize);

  columnVector *p = new columnVector(size);

  columnVector *ans = new columnVector(size);
  for(int i=0; i<size; i++) ans->push(i, 'a'+i);

  receiveMessage(dDist, H, r);


  //see how many blank columns we have
  int ret2=H->blankColumns();
  if(ret2!=-1) blankCols++; //printf("Blank column detected\n");

  //1. reorder (matrix and rows/columns in sync with received/source vectors)
  //return 0;
  
  //H->coordPrint();
  H->clearOpCounts();
  debugPrint("Triangulating\n");
  int sizeOfT = H->triangulate5(r, ans, rDist, -1);

  printCounts(H);

  debugPrint("Triangle size: %i\n", sizeOfT);


  avTri += sizeOfT;
  if(sizeOfT<minTri) minTri = sizeOfT;
  if(sizeOfT>maxTri) maxTri = sizeOfT;

  

  //check that the triangulate function preserved the original equations
  /*columnVector *r2 = new columnVector(size+spareSize);
  H->preMultiply(ans, r2);
  
  r2->prettyPrintTranspose();
  r->prettyPrintTranspose();
  delete r2;
  */
  //H->prettyPrint();

  if(sizeOfT==-1) { fprintf(stderr, "Fatal error\n");  H->prettyPrint(); return -1;}

  
  int ret = invert(H, p, r, ans);
  if(ret<0) {failed++; fail2-=ret;}

  
  printf("%i ", ret);
  printf("%i ", spareSize);
  printf("%i\n", sizeOfT);
  fflush(stdout);
  
  /*p->subtract(ans); */

  //printf("Answer is (differences):\n");

  //debugPrint("Answer is:\n");

  //p->prettyPrintTranspose();
  //printf("\n");

  //ans->prettyPrintTranspose();

  delete rDist; delete dDist;

  delete p;
  delete H; delete r;
  delete ans;

    seed++;
  }

  //printf("%i %i %f %f %i %i\n", blankCols, failed, (double)fail2/(double)reps2, (double)avTri/(double)(reps2*size), minTri, maxTri);
  printf("%i %f %i %i\n", failed, (double)avTri/(double)(reps2*size), minTri, maxTri);

  free(degi); free(degd);

  return 0;
}

