/* Copyright 2006 Oliver Madge
 * 
 * This file is part of RU-decoder.
 *
 * RU-decoder is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2 as published
 * by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program in the file COPYING; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301  USA
 *
 */

#include "degree.h"

degree::degree(int *vals, double *probs, int size, Random *rand) {
  //vals is a NULL terminated array of degrees
  this->r = rand;
  this->K = size;

  this->cache = (long double*)malloc(sizeof(long double)*this->K);
  
  for(int i=0; i<K; i++) {cache[i]=0.0;}

  int i=0;
  while(vals[i]!=0) {
    cache[vals[i]] = probs[i];
    i++;
  }
  pcache = NULL;
}

degree::~degree() {
  free(cache);
  if(pcache!=NULL) free(pcache);
}

uint16_t degree::getDegree() {

  double num = r->randomd();

  uint16_t i=1;
  num -= cache[i];
  while(num>0 && i<K-1){
    i++;
    num -= cache[i];
  }
  
  if(i==K) return 2;
  return i;
}

double degree::alpha_star() {
  long double alpha, diff, cutoff;
  int n=1000, prec=14;


  if(pcache==NULL) cache_poisson();
  //computer the mean of the distribution
  
  //divide and conquer for alpha
  
  
  alpha=0.5;

  diff = 0.25;

  cutoff = 0.0000001;
  
  int j=0;

  while(j<prec) {

    long double r = alpha;
 
    int i=0;
    while(i<n) {
      r = rho(1-r);
      //printf("r=%Le\n",r);
  
      r = alpha * lambda(1-r);
      //printf("r=%Le\n",r);
      i++;
    }
  
    printf("alpha=%Le, r=%Le\n", alpha, r);

    if(r>cutoff) { 
      alpha-=diff;
    }
    else { alpha+=diff; }
    
    diff*=0.5;
    j++;
  }

}

long double degree::rho(long double x) {
  long double ret=0.0;
  for(int i=0; i<maxp; i++) ret+=pcache[i]*pow(x, (i-1));
  return ret;
}

long double degree::lambda(long double x) {
  long double ret=0.0;
  for(int i=1; i<K; i++) ret+=cache[i]*pow(x, (i-1));
  return ret;
}

void degree::cache_poisson() {
  long double av=0.0;
  for(int i=0; i<this->K; i++) av+=i*cache[i];

  //printf("%Le\n", av);
  
  this->pcache = (long double*)malloc(sizeof(long double)*this->K);

  //av = 14.185776;

  pcache[0] =  exp(-av);
  maxp = (int)av*5;
  for(int k=1; k<maxp; k++) {
    //pcache[k] = premult * (pow(av, k)/ (long double)fact(k));
    pcache[k]=pcache[k-1]*av/k;
    //printf("%i %Le\n", k, pcache[k]);
  }
}

long degree::fact(int i) {
  if(i==0) return 1;
  return i*fact(i-1);
}


list<int>* degree::getPacketList(uint16_t deg, int list_size) {

  list<int> *output = new list<int>();
  long num;

  list<int>::iterator iter;

  if(deg>list_size) printf("Error, list too long\n");


  while(deg>0){
    num = r->randoml(0,list_size-1);
    iter = output->begin();
    while(iter!=output->end() && (*iter)<=num){
      iter++; num++;
    }
    output->insert(iter, num);
    list_size--;
    deg--;
  }

  return output;
}

