/* Copyright 2006 Oliver Madge
 * 
 * This file is part of RU-decoder.
 *
 * RU-decoder is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2 as published
 * by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program in the file COPYING; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301  USA
 *
 */

#include "dense.h"

//Constructor
denseMatrix::denseMatrix(int w, int h){
 
  this->width=w;
  this->height=h;
  
  int lim = this->width/BITS + 1;

  printf("lim=%i\n", lim);

  data = (unsigned int*)malloc(lim*this->height*sizeof(unsigned int));

  for(int i=0; i<lim*this->height; i++) data[i]=0;


}

//Destructor
denseMatrix::~denseMatrix(){
  free(data);
}

////////////////////////////////////////////////////////////////////////
//  Matrix manipulations
////////////////////////////////////////////////////////////////////////

void denseMatrix::setRow(int where, list<int> *input){

  list<int>::iterator iter = input->begin();

  int lim = this->width/BITS + 1;

  unsigned int tmp[lim], j=0;

  for(int i=0; i<lim; i++) {
    tmp[i]=0;
  }

  for(int i=0; i<this->width; i++) {
    if(*iter==i) {
      tmp[i/BITS]+=1<<((BITS-i)%BITS);
      iter++;
    }
  }
  
  for(int i=0; i<lim; i++) {
    printf("%u\n",tmp[i]);
    data[lim*where+i]=tmp[i];
  }
}

void denseMatrix::clear(){
  //clearRows();
  //clearColumns();
}

////////////////////////////////////////////////////////////////////////
//  Matrix inversion functions
////////////////////////////////////////////////////////////////////////

int denseMatrix::slackInvert(columnVector *a, columnVector *b){

  denseMatrix *m = new denseMatrix(this->width, this->width);
  columnVector *c = new columnVector(this->width);
  
  int lim = this->width/BITS+1;
  int j=0, ret=0, k, val;
  
  for(int i=0; i<this->height; i++) {
    //check if this row is blank
    //if so skip it
    val=-1;
    for(int k=0; k<lim; k++) {
      if(data[i*lim+k]!=0) {
        val=1;
        break;
      }
    }
    if(val==-1) {
      ret++;
      c->push(j, 0);
      continue;
    }
    
    if (val > j) { //then swap it with the current j
    //  
    }
    else if (val<j) {
      //add higher row to remove this
      i--;
      continue;
    }
    
    j++;
    if(j>=this->width) break;
    
  }
  
  if(j<this->width) ret=j-this->width;

  return ret;
}

/*
  //perform Gaussian elimination on this to get b=this^{-1}a
  //this can be over determined (height>width)

  //columnVector *c = new columnVector(this->width);

  //iteration i
  //find r = first non-zero column (first row element)
  //if r>i
    //swap column r with this column i
    //goto iteration i+1
  //else
    //add row r to row i
    //goto iteration i

  list<int>::const_iterator iter1;
  int val;
  int swri[this->height];
  int swrk[this->height];
  int j=0, k=0, ret=0;
  for(int i=0; i<this->height; i++) {
    iter1=(*this->rows)[i]->begin();
    if(iter1==(*this->rows)[i]->end()) {
      //printf("Row %i is blank\n", i);
      ret++;
      //delete the row
      c->push(j, 0);
      this->symbolOps++;
      continue;
    }
    
    val = *iter1;
    this->binaryOps++;
    if(val>j) {
      this->copyColumnsFromRows();
      this->swapColumns(val+1, j+1);
      m->swapColumns(val+1, j+1);
      this->copyRowsFromColumns();
      m->copyRowsFromColumns();
      swri[k] = val; swrk[k] = j;
      k++;
    }
    else if(val<j) {
      this->addRow(i, (*m->rows)[val]);
      c->push(j, c->get(j) ^ c->get(val));
      this->symbolOps++;
      i--;
      continue;
    }
    //else val==j
    //printf("Inverted row %i\n", i);

    c->push(j, c->get(j) ^ a->get(i));
    this->symbolOps++;

    iter1=(*this->rows)[i]->begin();
    while(iter1!=(*this->rows)[i]->end()) {
      m->insert(j+1, (*iter1)+1);
      this->binaryOps++;
      iter1++;
    }
    j++;
    if(j>=this->width) break;
  }

  if(j<this->width) ret=j-this->width;//printf("Underdetermined matrix\n");

  //now solve by back substitution
  m->pivotEliminate2(c, b);
  this->binaryOps+=m->getBinaryOpCount();
  this->symbolOps+=m->getSymbolOpCount();

  //swap back b
  for(int i=k-1; i>=0; i--) {
    b->swapRows(swri[i], swrk[i]);
    this->symbolOps++;
  }

  delete c;
  delete m;

  return ret;
}*/

////////////////////////////////////////////////////////////////////////
//  Display functions
////////////////////////////////////////////////////////////////////////

void denseMatrix::prettyPrint(){
  int lim = this->width/BITS+1;
  //loop over the rows
  for(int i=0; i<this->height; i++) {
    //list<int>::const_iterator iter=(*this->rows)[i]->begin();
    printf("|");
    for(int j=0; j < this->width; j++){
      
      if(data[i*lim+j/BITS]& (1<<abs((long long)(BITS-j)%BITS))){
        printf("1");
      }
      else printf("0");
    }
    printf("|\n");
  }
}


void denseMatrix::debugPrint(){
  int lim = this->width/BITS+1;
  //loop over the rows
  for(int i=0; i<this->height; i++) {
    printf("|");
    for(int j=0; j<lim; j++)
      printf(" %u ", data[lim*i+j]);
    printf("|\n");
  }
}

