/* Copyright 2006 Oliver Madge
 * 
 * This file is part of RU-decoder.
 *
 * RU-decoder is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2 as published
 * by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program in the file COPYING; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301  USA
 *
 */

#include "random.h"

Random::Random() {
  this->seeded = false;
  for (int i=0; i<5; i++) x[i]=0;
}

//FIXME: this isn't predictable across compilers due to rounding issues
//http://gcc.gnu.org/ml/gcc-patches/2001-03/msg00463.html

// returns a random number between 0 and 1:
double Random::randomd() {
  long double c;
  //if(!this->seeded) this->RandomInit(10);
  //splitting these up changes something
  c = (long double)2111111111.0 * (long double)this->x[3];
  c +=
      (double)1492.0 * (this->x[2]);
  c +=
      (double)1776.0 * (this->x[1]);
  c +=
      (double)5115.0 * (this->x[0]);
  c +=
      (double)this->x[4];

  this->x[3] = this->x[2];
  this->x[2] = this->x[1];
  this->x[1] = this->x[0];

  x[4] = floorl(c);
  x[0] = c - x[4];
  c = (65536.*65536.);
  c = 1.0 / c;
  x[4] = x[4] * c;
  return x[0];
}


// returns integer random number in desired interval:
long Random::randoml(long min, long max) {
  long iinterval = max - min + 1;
  if (iinterval <= 0) return 0x80000000; // error
  long i = long(iinterval * this->randomd());     // truncate
  if (i >= iinterval) i = iinterval-1;
  return min + i;
}


// this function initializes the random number generator:
void Random::RandomInit (uint32_t seed) {
  int i;
  uint32_t s = seed;
  // make random numbers and put them into the buffer
  for (i=0; i<5; i++) {
    s = s * 29943829 - 1;

    x[i] = s * (1./(65536.*65536.));
  }
  // randomize some more
  this->seeded = true;
  for (i=0; i<19; i++) randomd();
}

